@@ -56,14 +56,26 @@ CliqueManager::CliqueManager(int          const  rank,
                              cliqueMode_t const  cliqueMode) :
   m_rank(rank),
   m_numRanks(numRanks),
+  m_hash(0),
   m_cliqueMode(cliqueMode),
   m_opIndexHead(0),
   m_opIndexTail(0),
   m_init(false),
+  m_gcnArch(0),
+  m_allReduceByteLimit(0),
   m_pinnedCliquePtrs(NULL),
-  m_fineGrainBarrierMem(NULL)
-{
-}
+  m_gpuBarrierGlobalCount(NULL),
+  m_gpuBarrierGlobalSense(NULL),
+  m_gpuBarrierLocalSense(NULL),
+  m_cpuBarrierCount(NULL),
+  m_shmHandles(),
+  m_ipcHandleSendCache(),
+  m_ipcHandleRecvCache(),
+  m_sharedCpuMemory(),
+  m_sharedIpcHandle(),
+  m_fineGrainBarrierMem(NULL),
+  m_sharedBarrierCount(NULL)
+{}
 
 CliqueManager::~CliqueManager()
 {
@@ -84,6 +96,7 @@ void CliqueManager::CleanUp()
   if (m_cliqueMode == CLIQUE_SINGLE_NODE)
   {
     // Release caches
+    INFO(NCCL_COLL, "Rank %d deleting IPC caches", m_rank);
     if (m_ipcHandleSendCache) delete m_ipcHandleSendCache;
     if (m_ipcHandleRecvCache) delete m_ipcHandleRecvCache;
 
@@ -128,11 +141,6 @@ ncclResult_t CliqueManager::Init(ncclUniqueId const* commId, int suffix)
     WARN("Invalid rank specified.  Expected 0 <= %d < %d for CliqueManager", m_rank, m_numRanks);
     return ncclInvalidUsage;
   }
-  if (commId == NULL)
-  {
-    WARN("CommId should not be empty");
-    return ncclInvalidUsage;
-  }
 
   // For now, opt-into clique based kernels via RCCL_ENABLE_CLIQUE env var
   if (!rcclParamEnableClique())
@@ -182,7 +190,7 @@ ncclResult_t CliqueManager::Init(ncclUniqueId const* commId, int suffix)
       hipIpcMemHandle_t handle;
       // Allocate fine-grained device memory on rank 0 and get IPC handle for it
       // Re-usable barrier consists of (globalCount / globalSense) pair of integers
-      NCCLCHECKGOTO(ncclCudaCalloc(&m_fineGrainBarrierMem, NCCL_MAX_OPS * 2 * sizeof(int), true), res, dropback);
+      NCCLCHECKGOTO(ncclCudaCalloc(&m_fineGrainBarrierMem, NCCL_MAX_OPS * 2 * sizeof(int), nullptr, true), res, dropback);
       if (hipIpcGetMemHandle(&handle, m_fineGrainBarrierMem) != hipSuccess)
       {
         WARN("Unable to get IPC handle for barrier memory");
@@ -220,7 +228,7 @@ ncclResult_t CliqueManager::Init(ncclUniqueId const* commId, int suffix)
     // First rank prepares fine-grained memory shared across ranks used for the two barrier variables
     if (m_rank == 0)
     {
-      NCCLCHECKGOTO(ncclCudaCalloc(&m_staticGpuBarrierMem, NCCL_MAX_OPS * 2 * sizeof(int), true), res, dropback);
+      NCCLCHECKGOTO(ncclCudaCalloc(&m_staticGpuBarrierMem, NCCL_MAX_OPS * 2 * sizeof(int), nullptr, true), res, dropback);
       // Prepare all barriers
       for (int opIndex = 0; opIndex < NCCL_MAX_OPS; opIndex++)
       {
@@ -261,7 +269,7 @@ void CliqueManager::SetByteLimits()
     switch (m_gcnArch)
     {
     case 906: m_allReduceByteLimit =  16777216; break;
-    case 908: m_allReduceByteLimit = 536870912; break;
+    case 908: m_allReduceByteLimit =   8388608; break;
     default:  m_allReduceByteLimit =  16777216; break;
     }
   }
@@ -274,9 +282,9 @@ bool CliqueManager::IsSupported(ncclFunc_t const coll,
 {
   if (m_cliqueMode == CLIQUE_DISABLED) return false;
 
-  // Filter based on total input size for each collective type
+  // Filter based on total input size for each collective type and ops sum/prod/min/max
   size_t totalBytes = count * ncclTypeSize(datatype);
-  if (coll == ncclFuncAllReduce && (totalBytes <= m_allReduceByteLimit)) return true;
+  if (coll == ncclFuncAllReduce && (totalBytes <= m_allReduceByteLimit) && op < ncclAvg) return true;
   return false;
 }
 
@@ -350,7 +358,7 @@ ncclResult_t CliqueManager::GetNumChannelsToUse(ncclFunc_t const coll,
                                                 ncclDataType_t const datatype,
                                                 ncclRedOp_t const op,
                                                 int const totalNumChannels,
-                                                uint8_t* numChannelstoUse)
+                                                uint8_t* numChannelstoUse) const
 {
   size_t const totalBytes = count * ncclTypeSize(datatype);
   *numChannelstoUse = 1;
@@ -367,8 +375,10 @@ ncclResult_t CliqueManager::GetNumChannelsToUse(ncclFunc_t const coll,
         else                            numChannels =  2;
         break;
       case 908:
-        if      (totalBytes <=  262144) numChannels =  4;
-        else                            numChannels = 14;
+        if      (totalBytes <=  131072) numChannels =  2;
+        else if (totalBytes <=  524288) numChannels =  6;
+        else if (totalBytes <= 1048576) numChannels = 13;
+        else                            numChannels = 16;
         break;
       case 910:
         if      (totalBytes <=  262144) numChannels =  4;
@@ -465,20 +475,6 @@ ncclResult_t CliqueManager::WaitForPointers(ncclWorkElem* args)
   return ncclSuccess;
 }
 
-std::string HandleToString(hipIpcMemHandle_t handle)
-{
-  char mapping[17] = "0123456789ABCDEF";
-  std::string result;
-  for (int i = 0; i < 4; i++)
-  {
-    unsigned char val = (unsigned char)handle.reserved[i];
-    result += mapping[val / 16];
-    result += mapping[val % 16];
-  }
-  return result;
-}
-
-
 ncclResult_t CliqueManager::CheckCacheForPtr(void* devPtr,
                                              NcclIpcHandleSendCache* cache,
                                              int rank,
@@ -499,18 +495,24 @@ ncclResult_t CliqueManager::CheckCacheForPtr(void* devPtr,
   uint64_t realAddr = (uint64_t)devPtr;
   handlePair->second = realAddr - baseAddr;
 
+  CUDACHECK(hipIpcGetMemHandle(&handlePair->first, (void*)baseAddr));
+
+  /* Disabling cache until proper deallocation methods are available
   // IPC handles are only supported for base address pointers
   NcclIpcHandleSendCache::iterator it = cache->find(baseAddr);
 
    if (it == cache->end())
    {
+     INFO(NCCL_COLL, "Rank %d searching IPC handle cache for %p (not found)", rank, devPtr);
      CUDACHECK(hipIpcGetMemHandle(&handlePair->first, (void*)baseAddr));
      cache->insert(baseAddr, handlePair->first);
    }
    else
    {
+     INFO(NCCL_COLL, "Rank %d searching IPC handle cache for %p (found!)", rank, devPtr);
      handlePair->first = (it->second).first;
    }
+  */
    return ncclSuccess;
 }
 
@@ -518,10 +520,16 @@ ncclResult_t CliqueManager::CheckCacheForHandle(std::pair<hipIpcMemHandle_t, siz
                                                 NcclIpcHandleRecvCache* cache,
                                                 void** ptr)
 {
+  // Until proper deallocation hooks are implemented, receive cache can not be used
+  // Handles will need to be extract each time
+  void* baseAddr;
+  CUDACHECK(hipIpcOpenMemHandle(&baseAddr, handlePair.first, hipIpcMemLazyEnablePeerAccess));
+
+  /*
   NcclIpcHandleRecvCache::iterator it = cache->find(handlePair.first);
 
   // Get base address pointer from cache if it exists
-  void* baseAddr;
+
   if (it == cache->end())
   {
     CUDACHECK(hipIpcOpenMemHandle(&baseAddr, handlePair.first, hipIpcMemLazyEnablePeerAccess));
@@ -531,6 +539,7 @@ ncclResult_t CliqueManager::CheckCacheForHandle(std::pair<hipIpcMemHandle_t, siz
   {
     baseAddr = (it->second).first;
   }
+  */
 
   // Modify base address pointer with offset
   uint64_t realAddr = (uint64_t)baseAddr + handlePair.second;
