@@ -9,29 +9,37 @@
 #include "nccl_net.h"
 #include <stdlib.h>
 #include <stdarg.h>
+#include <sys/syscall.h>
 
 int ncclDebugLevel = -1;
+static int pid = -1;
+static char hostname[1024];
 thread_local int ncclDebugNoWarn = 0;
+char ncclLastError[1024] = ""; // Global string for the last error in human readable form
 uint64_t ncclDebugMask = NCCL_INIT; // Default debug sub-system mask is INIT
 FILE *ncclDebugFile = stdout;
 pthread_mutex_t ncclDebugLock = PTHREAD_MUTEX_INITIALIZER;
+std::chrono::steady_clock::time_point ncclEpoch;
+
+static __thread int tid = -1;
 
 void ncclDebugInit() {
   pthread_mutex_lock(&ncclDebugLock);
   if (ncclDebugLevel != -1) { pthread_mutex_unlock(&ncclDebugLock); return; }
   const char* nccl_debug = getenv("NCCL_DEBUG");
+  int tempNcclDebugLevel = -1;
   if (nccl_debug == NULL) {
-    ncclDebugLevel = NCCL_LOG_NONE;
+    tempNcclDebugLevel = NCCL_LOG_NONE;
   } else if (strcasecmp(nccl_debug, "VERSION") == 0) {
-    ncclDebugLevel = NCCL_LOG_VERSION;
+    tempNcclDebugLevel = NCCL_LOG_VERSION;
   } else if (strcasecmp(nccl_debug, "WARN") == 0) {
-    ncclDebugLevel = NCCL_LOG_WARN;
+    tempNcclDebugLevel = NCCL_LOG_WARN;
   } else if (strcasecmp(nccl_debug, "INFO") == 0) {
-    ncclDebugLevel = NCCL_LOG_INFO;
+    tempNcclDebugLevel = NCCL_LOG_INFO;
   } else if (strcasecmp(nccl_debug, "ABORT") == 0) {
-    ncclDebugLevel = NCCL_LOG_ABORT;
+    tempNcclDebugLevel = NCCL_LOG_ABORT;
   } else if (strcasecmp(nccl_debug, "TRACE") == 0) {
-    ncclDebugLevel = NCCL_LOG_TRACE;
+    tempNcclDebugLevel = NCCL_LOG_TRACE;
   }
 
   /* Parse the NCCL_DEBUG_SUBSYS env var
@@ -65,6 +73,8 @@ void ncclDebugInit() {
         mask = NCCL_ENV;
       } else if (strcasecmp(subsys, "ALLOC") == 0) {
         mask = NCCL_ALLOC;
+      } else if (strcasecmp(subsys, "CALL") == 0) {
+        mask = NCCL_CALL;
       } else if (strcasecmp(subsys, "ALL") == 0) {
         mask = NCCL_ALL;
       }
@@ -76,12 +86,16 @@ void ncclDebugInit() {
     free(ncclDebugSubsys);
   }
 
+  // Cache pid and hostname
+  getHostName(hostname, 1024, '.');
+  pid = getpid();
+
   /* Parse and expand the NCCL_DEBUG_FILE path and
    * then create the debug file. But don't bother unless the
    * NCCL_DEBUG level is > VERSION
    */
   const char* ncclDebugFileEnv = getenv("NCCL_DEBUG_FILE");
-  if (ncclDebugLevel > NCCL_LOG_VERSION && ncclDebugFileEnv != NULL) {
+  if (tempNcclDebugLevel > NCCL_LOG_VERSION && ncclDebugFileEnv != NULL) {
     int c = 0;
     char debugFn[PATH_MAX+1] = "";
     char *dfn = debugFn;
@@ -95,12 +109,10 @@ void ncclDebugInit() {
           *dfn++ = '%';
           break;
         case 'h': // %h = hostname
-          char hostname[1024];
-          getHostName(hostname, 1024, '.');
           dfn += snprintf(dfn, PATH_MAX, "%s", hostname);
           break;
         case 'p': // %p = pid
-          dfn += snprintf(dfn, PATH_MAX, "%d", getpid());
+          dfn += snprintf(dfn, PATH_MAX, "%d", pid);
           break;
         default: // Echo everything we don't understand
           *dfn++ = '%';
@@ -111,15 +123,15 @@ void ncclDebugInit() {
     *dfn = '\0';
     if (debugFn[0] != '\0') {
       FILE *file = fopen(debugFn, "w");
-      if (file != NULL) {
+      if (file != nullptr) {
+        setbuf(file, nullptr); // disable buffering
         ncclDebugFile = file;
       }
     }
   }
 
-#ifdef ENABLE_TRACE
-  ncclEpoch = std::chrono::high_resolution_clock::now();
-#endif
+  ncclEpoch = std::chrono::steady_clock::now();
+  __atomic_store_n(&ncclDebugLevel, tempNcclDebugLevel, __ATOMIC_RELEASE);
   pthread_mutex_unlock(&ncclDebugLock);
 }
 
@@ -128,45 +140,53 @@ void ncclDebugInit() {
  * they can share the debugging mechanisms and output files
  */
 void ncclDebugLog(ncclDebugLogLevel level, unsigned long flags, const char *filefunc, int line, const char *fmt, ...) {
-  if (ncclDebugLevel == -1) ncclDebugInit();
+  if (__atomic_load_n(&ncclDebugLevel, __ATOMIC_ACQUIRE) == -1) ncclDebugInit();
   if (ncclDebugNoWarn != 0 && level == NCCL_LOG_WARN) { level = NCCL_LOG_INFO; flags = ncclDebugNoWarn; }
+
+  // Save the last error (WARN) as a human readable string
+  if (level == NCCL_LOG_WARN) {
+    pthread_mutex_lock(&ncclDebugLock);
+    va_list vargs;
+    va_start(vargs, fmt);
+    (void) vsnprintf(ncclLastError, sizeof(ncclLastError), fmt, vargs);
+    va_end(vargs);
+    pthread_mutex_unlock(&ncclDebugLock);
+  }
   if (ncclDebugLevel < level || ((flags & ncclDebugMask) == 0)) return;
 
-  // Gather the rank information. This can take > 1us so we want to make sure
-  // we only do it when needed.
-  char hostname[1024];
-  getHostName(hostname, 1024, '.');
+  if (tid == -1) {
+    tid = syscall(SYS_gettid);
+  }
+
   int cudaDev;
-  hipGetDevice(&cudaDev);
-  int pid = getpid();
-  int tid = syscall(SYS_gettid);
+  if (!(level == NCCL_LOG_TRACE && flags == NCCL_CALL)) {
+    hipGetDevice(&cudaDev);
+  }
 
   char buffer[1024];
   size_t len = 0;
-  pthread_mutex_lock(&ncclDebugLock);
-  if (level == NCCL_LOG_WARN)
-    len = snprintf(buffer, sizeof(buffer),
-        "\n%s:%d:%d [%d] %s:%d NCCL WARN ", hostname, pid, tid, cudaDev, filefunc, line);
-  else if (level == NCCL_LOG_INFO)
-    len = snprintf(buffer, sizeof(buffer),
-        "%s:%d:%d [%d] NCCL INFO ", hostname, pid, tid, cudaDev);
-#ifdef ENABLE_TRACE
-  else if (level == NCCL_LOG_TRACE) {
-    auto delta = std::chrono::high_resolution_clock::now() - ncclEpoch;
+  if (level == NCCL_LOG_WARN) {
+    len = snprintf(buffer, sizeof(buffer), "\n%s:%d:%d [%d] %s:%d NCCL WARN ",
+                   hostname, pid, tid, cudaDev, filefunc, line);
+  } else if (level == NCCL_LOG_INFO) {
+    len = snprintf(buffer, sizeof(buffer), "%s:%d:%d [%d] NCCL INFO ", hostname, pid, tid, cudaDev);
+  } else if (level == NCCL_LOG_TRACE && flags == NCCL_CALL) {
+    len = snprintf(buffer, sizeof(buffer), "%s:%d:%d NCCL CALL ", hostname, pid, tid);
+  } else if (level == NCCL_LOG_TRACE) {
+    auto delta = std::chrono::steady_clock::now() - ncclEpoch;
     double timestamp = std::chrono::duration_cast<std::chrono::duration<double>>(delta).count()*1000;
-    len = snprintf(buffer, sizeof(buffer),
-        "%s:%d:%d [%d] %f %s:%d NCCL TRACE ", hostname, pid, tid, cudaDev, timestamp, filefunc, line);
+    len = snprintf(buffer, sizeof(buffer), "%s:%d:%d [%d] %f %s:%d NCCL TRACE ",
+                   hostname, pid, tid, cudaDev, timestamp, filefunc, line);
   }
-#endif
+
   if (len) {
     va_list vargs;
     va_start(vargs, fmt);
-    (void) vsnprintf(buffer+len, sizeof(buffer)-len, fmt, vargs);
+    len += vsnprintf(buffer+len, sizeof(buffer)-len, fmt, vargs);
     va_end(vargs);
-    fprintf(ncclDebugFile,"%s\n", buffer);
-    fflush(ncclDebugFile);
+    buffer[len++] = '\n';
+    fwrite(buffer, 1, len, ncclDebugFile);
   }
-  pthread_mutex_unlock(&ncclDebugLock);
 }
 
 NCCL_PARAM(SetThreadName, "SET_THREAD_NAME", 0);
