@@ -37,7 +37,9 @@ static __inline__ int ncclTypeSize(ncclDataType_t type) {
     case ncclUint8:
       return 1;
     case ncclFloat16:
+#if defined(RCCL_BFLOAT16)
     case ncclBfloat16:
+#endif
       return 2;
     case ncclInt32:
     case ncclUint32:
@@ -54,6 +56,7 @@ static __inline__ int ncclTypeSize(ncclDataType_t type) {
 
 #include "debug.h"
 #include "checks.h"
+#include "rocmwrap.h"
 #include "alloc.h"
 #include "utils.h"
 #include "param.h"
