@@ -13,7 +13,7 @@ namespace RcclUnitTesting
 
     // Configuration
     std::vector<ncclFunc_t>     const funcTypes      = {ncclCollAllReduce};
-    std::vector<ncclDataType_t> const dataTypes      = {ncclFloat32, ncclFloat64, ncclBfloat16};
+    std::vector<ncclDataType_t> const dataTypes      = {ncclFloat16, ncclFloat32, ncclFloat64, ncclBfloat16};
     std::vector<ncclRedOp_t>    const redOps         = {ncclSum, ncclMin};
     std::vector<int>            const roots          = {0};
     std::vector<int>            const numElements    = {1048576, 53327, 1024};
