@@ -13,10 +13,10 @@ namespace RcclUnitTesting
 
     // Configuration
     std::vector<ncclFunc_t>     const funcTypes      = {ncclCollReduceScatter};
-    std::vector<ncclDataType_t> const dataTypes      = {ncclFloat32, ncclFloat64, ncclBfloat16};
+    std::vector<ncclDataType_t> const dataTypes      = {ncclFloat16, ncclFloat32, ncclFloat64, ncclBfloat16};
     std::vector<ncclRedOp_t>    const redOps         = {ncclMin, ncclMax, ncclAvg};
     std::vector<int>            const roots          = {0};
-    std::vector<int>            const numElements    = {1048576, 53327, 1024};
+    std::vector<int>            const numElements    = {1048576, 53327, 5461, 1024};
     std::vector<bool>           const inPlaceList    = {false};
     std::vector<bool>           const managedMemList = {false};
 
